#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["signal_batch_processed", "signal_batch_started", "signal_batch_lag"]

from queue import (
    Full,
    Queue,
)
from time import monotonic
from typing import Optional

from neptune.common.warnings import (
    NeptuneWarning,
    warn_once,
)
from neptune.internal.signals_processing.signals import (
    BatchLagSignal,
    BatchProcessedSignal,
    BatchStartedSignal,
    Signal,
)


def signal(*, queue: "Queue[Signal]", obj: "Signal") -> None:
    try:
        queue.put_nowait(item=obj)
    except Full:
        warn_once("Signal queue is full. Some signals will be lost.", exception=NeptuneWarning)


def signal_batch_started(*, queue: "Queue[Signal]", occured_at: Optional[float] = None) -> None:
    signal(queue=queue, obj=BatchStartedSignal(occured_at=occured_at or monotonic()))


def signal_batch_processed(*, queue: "Queue[Signal]", occured_at: Optional[float] = None) -> None:
    signal(queue=queue, obj=BatchProcessedSignal(occured_at=occured_at or monotonic()))


def signal_batch_lag(*, queue: "Queue[Signal]", lag: float, occured_at: Optional[float] = None) -> None:
    signal(queue=queue, obj=BatchLagSignal(occured_at=occured_at or monotonic(), lag=lag))



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["load_extensions"]

import sys
from typing import (
    Callable,
    List,
    Tuple,
)

if sys.version_info >= (3, 8):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points

from neptune.common.warnings import (
    NeptuneWarning,
    warn_once,
)


def get_entry_points(name: str) -> List[Tuple[str, Callable[[], None]]]:
    if (3, 8) <= sys.version_info < (3, 10):
        return [(entry_point.name, entry_point.load()) for entry_point in entry_points().get(name, tuple())]
    return [
        (entry_point.name, entry_point.load())  # type: ignore[unused-ignore, attr-defined]
        for entry_point in entry_points(group=name)  # type: ignore[unused-ignore, call-arg]
    ]


def load_extensions() -> None:
    for entry_point_name, loaded_extension in get_entry_points(name="neptune.extensions"):
        try:
            _ = loaded_extension()
        except Exception as e:
            warn_once(
                message=f"Failed to load neptune extension `{entry_point_name}` with exception: {e}",
                exception=NeptuneWarning,
            )



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_single_page", "iter_over_pages"]

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
)

from bravado.client import construct_request  # type: ignore
from bravado.config import RequestConfig  # type: ignore

from neptune.internal.backends.api_model import (
    AttributeType,
    AttributeWithProperties,
    LeaderboardEntry,
)
from neptune.internal.backends.hosted_client import DEFAULT_REQUEST_KWARGS
from neptune.internal.backends.nql import (
    NQLAggregator,
    NQLAttributeOperator,
    NQLAttributeType,
    NQLEmptyQuery,
    NQLQuery,
    NQLQueryAggregate,
    NQLQueryAttribute,
)
from neptune.internal.init.parameters import MAX_SERVER_OFFSET

if TYPE_CHECKING:
    from neptune.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
    from neptune.internal.id_formats import UniqueId


SUPPORTED_ATTRIBUTE_TYPES = {item.value for item in AttributeType}


def get_single_page(
    *,
    client: "SwaggerClientWrapper",
    project_id: "UniqueId",
    attributes_filter: Dict[str, Any],
    limit: int,
    offset: int,
    sort_by: Optional[str] = None,
    types: Optional[Iterable[str]] = None,
    query: Optional["NQLQuery"] = None,
    searching_after: Optional[str] = None,
) -> List[Any]:
    normalized_query = query or NQLEmptyQuery()
    if sort_by and searching_after:
        sort_by_as_nql = NQLQueryAttribute(
            name=sort_by,
            type=NQLAttributeType.STRING,
            operator=NQLAttributeOperator.GREATER_THAN,
            value=searching_after,
        )

        if not isinstance(normalized_query, NQLEmptyQuery):
            normalized_query = NQLQueryAggregate(items=[normalized_query, sort_by_as_nql], aggregator=NQLAggregator.AND)
        else:
            normalized_query = sort_by_as_nql

    sorting = (
        {
            "sorting": {
                "dir": "ascending",
                "aggregationMode": "none",
                "sortBy": {"name": sort_by, "type": "string"},
            }
        }
        if sort_by
        else {}
    )

    params = {
        "projectIdentifier": project_id,
        "type": types,
        "params": {
            **sorting,
            **attributes_filter,
            "query": {"query": str(normalized_query)},
            "pagination": {"limit": limit, "offset": offset},
        },
    }

    request_options = DEFAULT_REQUEST_KWARGS.get("_request_options", {})
    request_config = RequestConfig(request_options, True)
    request_params = construct_request(client.api.searchLeaderboardEntries, request_options, **params)

    http_client = client.swagger_spec.http_client

    result = (
        http_client.request(request_params, operation=None, request_config=request_config)
        .response()
        .incoming_response.json()
    )

    return list(map(to_leaderboard_entry, result.get("entries", [])))


def to_leaderboard_entry(entry: Dict[str, Any]) -> LeaderboardEntry:
    return LeaderboardEntry(
        id=entry["experimentId"],
        attributes=[
            AttributeWithProperties(
                path=attr["name"],
                type=AttributeType(attr["type"]),
                properties=attr.__getitem__(f"{attr['type']}Properties"),
            )
            for attr in entry["attributes"]
            if attr["type"] in SUPPORTED_ATTRIBUTE_TYPES
        ],
    )


def find_attribute(*, entry: LeaderboardEntry, path: str) -> Optional[AttributeWithProperties]:
    return next((attr for attr in entry.attributes if attr.path == path), None)


def iter_over_pages(
    *,
    step_size: int,
    sort_by: str = "sys/id",
    max_offset: int = MAX_SERVER_OFFSET,
    **kwargs: Any,
) -> Generator[Any, None, None]:
    searching_after = None
    last_page = None

    while True:
        if last_page:
            page_attribute = find_attribute(entry=last_page[-1], path=sort_by)

            if not page_attribute:
                raise ValueError(f"Cannot find attribute {sort_by} in last page")

            searching_after = page_attribute.properties["value"]

        for offset in range(0, max_offset, step_size):
            page = get_single_page(
                limit=min(step_size, max_offset - offset),
                offset=offset,
                sort_by=sort_by,
                searching_after=searching_after,
                **kwargs,
            )

            if not page:
                return

            yield from page

            last_page = page



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["MetadataFile"]

import json
import os
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Dict,
    Optional,
    Type,
)


class MetadataFile:
    METADATA_FILE: str = "metadata.json"

    def __init__(self, data_path: Path, metadata: Optional[Dict[str, Any]] = None):
        # initialize directory
        os.makedirs(data_path, exist_ok=True)

        self._metadata_path: Path = (data_path / MetadataFile.METADATA_FILE).resolve(strict=False)
        self._data: Dict[str, Any] = self._read_or_default()

        if metadata:
            for key, value in metadata.items():
                self.__setitem__(key, value)
            self.flush()

    def __getitem__(self, item: str) -> Any:
        return self._data[item]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def flush(self) -> None:
        with open(self._metadata_path, "w") as handler:
            json.dump(self._data, handler, indent=2)

    def _read_or_default(self) -> Dict[str, Any]:
        if self._metadata_path.exists():
            with open(self._metadata_path, "r") as handler:
                data: Dict[str, Any] = json.load(handler)
                return data

        return dict()

    def close(self) -> None:
        self.flush()

    def cleanup(self) -> None:
        try:
            os.remove(self._metadata_path)
        except OSError:
            pass

    def __enter__(self) -> "MetadataFile":
        return self

    def __exit__(
        self,
        exc_type: Type[Optional[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["common_metadata"]

import datetime
import platform
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
)

if TYPE_CHECKING:
    from neptune.internal.container_type import ContainerType
    from neptune.internal.id_formats import UniqueId


def get_neptune_version() -> str:
    from neptune.version import __version__ as neptune_version

    return neptune_version


def common_metadata(mode: str, container_id: "UniqueId", container_type: "ContainerType") -> Dict[str, Any]:
    return {
        "mode": mode,
        "containerId": container_id,
        "containerType": container_type,
        "structureVersion": 1,
        "os": platform.platform(),
        "pythonVersion": sys.version,
        "neptuneClientVersion": get_neptune_version(),
        "createdAt": datetime.datetime.utcnow().isoformat(),
    }



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["ensure_disk_not_overutilize"]


import os
from abc import (
    ABC,
    abstractmethod,
)
from functools import wraps
from typing import (
    Any,
    Callable,
    Optional,
)

import psutil
from psutil import Error

from neptune.common.warnings import (
    NeptuneWarning,
    warn_once,
)
from neptune.constants import NEPTUNE_DATA_DIRECTORY
from neptune.envs import (
    NEPTUNE_MAX_DISK_USAGE,
    NEPTUNE_RAISE_ERROR_ON_DISK_USAGE_EXCEEDED,
)
from neptune.exceptions import NeptuneMaxDiskUtilizationExceeded


def get_neptune_data_directory() -> str:
    return os.getenv("NEPTUNE_DATA_DIRECTORY", NEPTUNE_DATA_DIRECTORY)


def get_disk_utilization_percent(path: Optional[str] = None) -> Optional[float]:
    try:
        if path is None:
            path = get_neptune_data_directory()

        return float(psutil.disk_usage(path).percent)
    except (ValueError, TypeError, Error):
        return None


def get_max_disk_utilization_from_env() -> Optional[float]:
    env_limit_disk_utilization = os.getenv(NEPTUNE_MAX_DISK_USAGE)

    if env_limit_disk_utilization is None:
        return None

    try:
        limit_disk_utilization = float(env_limit_disk_utilization)
        if limit_disk_utilization <= 0 or limit_disk_utilization > 100:
            raise ValueError

        return limit_disk_utilization
    except (ValueError, TypeError):
        warn_once(
            f"Provided invalid value of '{NEPTUNE_MAX_DISK_USAGE}': '{env_limit_disk_utilization}'. "
            "Check of disk utilization will not be applied.",
            exception=NeptuneWarning,
        )
        return None


class DiskUtilizationErrorHandlerTemplate(ABC):
    def __init__(self, max_disk_utilization: Optional[float], func: Callable[..., None], *args: Any, **kwargs: Any):
        self.max_disk_utilization = max_disk_utilization
        self.func = func
        self.args = args
        self.kwargs = kwargs

    @abstractmethod
    def handle_limit_not_set(self) -> None:
        ...  # pragma: no cover

    @abstractmethod
    def handle_utilization_calculation_error(self) -> None:
        ...  # pragma: no cover

    @abstractmethod
    def handle_limit_not_exceeded(self) -> None:
        ...  # pragma: no cover

    @abstractmethod
    def handle_limit_exceeded(self, current_utilization: float) -> None:
        ...  # pragma: no cover

    def run(self) -> None:
        if not self.max_disk_utilization:
            return self.handle_limit_not_set()

        current_utilization = get_disk_utilization_percent()

        if current_utilization is None:
            return self.handle_utilization_calculation_error()

        if current_utilization < self.max_disk_utilization:
            return self.handle_limit_not_exceeded()

        self.handle_limit_exceeded(current_utilization)


class NonRaisingErrorHandler(DiskUtilizationErrorHandlerTemplate):
    DISK_ISSUE_MSG = "Encountered disk issue. Neptune will not save your data."

    def handle_limit_not_set(self) -> None:
        try:
            return self.func(*self.args, **self.kwargs)
        except (OSError, Error):
            warn_once(self.DISK_ISSUE_MSG, exception=NeptuneWarning)

    def handle_utilization_calculation_error(self) -> None:
        try:
            return self.func(*self.args, **self.kwargs)
        except (OSError, Error):
            warn_once(self.DISK_ISSUE_MSG, exception=NeptuneWarning)

    def handle_limit_not_exceeded(self) -> None:
        try:
            return self.func(*self.args, **self.kwargs)
        except (OSError, Error):
            warn_once(self.DISK_ISSUE_MSG, exception=NeptuneWarning)

    def handle_limit_exceeded(self, current_utilization: float) -> None:
        warn_once(
            f"Disk usage is at {current_utilization}%, which exceeds the maximum allowed utilization "
            + f"of {self.max_disk_utilization}%. Neptune will not save your data.",
            exception=NeptuneWarning,
        )


class RaisingErrorHandler(DiskUtilizationErrorHandlerTemplate):
    def handle_limit_not_set(self) -> None:
        return self.func(*self.args, **self.kwargs)

    def handle_utilization_calculation_error(self) -> None:
        return self.func(*self.args, **self.kwargs)

    def handle_limit_not_exceeded(self) -> None:
        return self.func(*self.args, **self.kwargs)

    def handle_limit_exceeded(self, current_utilization: float) -> None:
        if isinstance(self.max_disk_utilization, float):

            raise NeptuneMaxDiskUtilizationExceeded(
                disk_utilization=current_utilization,
                utilization_limit=self.max_disk_utilization,
            )


def ensure_disk_not_overutilize(func: Callable[..., None]) -> Callable[..., None]:
    raising_on_disk_issue = os.getenv(NEPTUNE_RAISE_ERROR_ON_DISK_USAGE_EXCEEDED, "True").lower() in ("true", "t", "1")
    max_disk_utilization = get_max_disk_utilization_from_env()

    error_handler = RaisingErrorHandler if raising_on_disk_issue else NonRaisingErrorHandler

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        error_handler(max_disk_utilization, func, *args, **kwargs).run()

    return wrapper



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from dataclasses import dataclass


@dataclass(frozen=True)
class MultipartConfig:
    min_chunk_size: int
    max_chunk_size: int
    max_chunk_count: int
    max_single_part_size: int

    @staticmethod
    def get_default() -> "MultipartConfig":
        return MultipartConfig(
            min_chunk_size=5242880,
            max_chunk_size=1073741824,
            max_chunk_count=1000,
            max_single_part_size=5242880,
        )
